/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.consume;

import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_CURRENT_DATA_DELIMITER_TOKEN;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_CURRENT_DELIMITER;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_INITIAL_DELIMITER;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_XQUERY;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_PREVIOUS_DATA_DELIMITER_TOKEN;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.sql.DataSource;

import org.ow2.petals.bc.xquare.XQuareBCException;
import org.ow2.petals.bc.xquare.util.XQuareUtils;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.w3c.dom.Document;
import org.xquark.bridge.XQBridge;

/**
 * <p>
 * Implements a strategy for detecting "new data", getting it (as XML) and
 * updating what data is known, by monitoring a given field as its "new data
 * delimiter".
 * </p>
 * 
 * <p>
 * What it actually does is first executing two SQL queries to get the "new",
 * current data delimiter value and the "old", known data delimiter value. Then
 * it executes an XQuery , after having the "##OLD_DATA_DELIMITER##" and
 * "##NEW_DATA_DELIMITER##" expressions replaced accordingly within its XQuery
 * code. Finally it executes an SQL insertion to update the "old", known data
 * delimiter to the current "new" one.
 * </p>
 * 
 * <p>
 * NB. All These SQL and XQuery queries, insertions and initial value are
 * configured in the extensions of the SU descriptor.
 * 
 * NB. Must be instantiated for each new data detection. NB. This is the only
 * implementation of NewDataStrategy for now.
 * 
 * @version $Rev: 8909 $ $Date: 2008-08-21 18:08:20 +0200 (Thu, 21 Aug 2008) $
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 * 
 */
public abstract class DelimiterNewDataStrategy implements NewDataStrategy {

    private Logger logger;

    protected String previousDataDelimiterSelect;

    protected String currentDataDelimiterSelect;

    protected String initialDelimiter;
    
    protected String updateDelimiters;

    protected String newDataXquery;

    protected ArrayList<String> newDataDelimiterUpdates = new ArrayList<String>();

    protected String oldDataDelimiter;

    protected String newDataDelimiter;

    private XQuareBCExternalListener processor;

    public ConfigurationExtensions extensions;

    protected XQBridge xqBridge;
    
    protected DataSource dataSource;
    
    /**
     * Creates a new DelimiterNewDataStrategy
     * 
     * @param processor
     * @param conn
     * @param extensions
     * @param listenerPropPrefix
     * @param logger
     * @throws XQuareBCException
     */
    public DelimiterNewDataStrategy() {
    }
    
    public void init(XQuareBCExternalListener processor, XQBridge xqBridge,
            ConfigurationExtensions extensions, Logger logger) throws XQuareBCException {
        this.xqBridge = xqBridge;
        this.extensions = extensions;
        this.logger = logger;

        // for query :
        this.processor = processor;

        // init from props :
        this.currentDataDelimiterSelect = extensions.get(XQUARE_DATALISTENER_CURRENT_DELIMITER);
        this.initialDelimiter = extensions.get(XQUARE_DATALISTENER_INITIAL_DELIMITER);
        this.newDataXquery = extensions.get(XQUARE_DATALISTENER_XQUERY);
        
        // creating simple pooled datasource
        // (using BasicDataSource.createDataSource() which sets up a default pool conf)
        
        //TODO reuse DataSources manager
        dataSource = null;//XQuareUtils.newBasicDataSource(extensions);
    }
    
    public abstract String getOldDataDelimiter() throws Exception;
    
    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.bc.xquarebc.listeners.NewDataStrategy#detectNewData()
     */
    public boolean detectNewData() throws Exception {
        // getting the old data delimiter :
        oldDataDelimiter = getOldDataDelimiter();
        if (oldDataDelimiter == null) {
            oldDataDelimiter = initialDelimiter;
        }

        // getting the new data delimiter :
        Connection connection = null;
        Statement getNewMaxIdStatement = null;
        try {
        	connection = dataSource.getConnection();
        	getNewMaxIdStatement = connection.createStatement();
        	getNewMaxIdStatement.execute(this.currentDataDelimiterSelect);
        	
	        ResultSet getNewMaxIdRs = getNewMaxIdStatement.getResultSet();
	        if (getNewMaxIdRs.next()) {
	            newDataDelimiter = getNewMaxIdRs.getString(1);
	        }
	        
	        // there is new data if they are different
	        return (newDataDelimiter != null && !newDataDelimiter.equals(oldDataDelimiter));

        } catch (Exception e) {
            logger.log(Level.FINE, "Error while detecting new data ; not closing connection though", e);
            throw e;
        } finally {
        	if (getNewMaxIdStatement != null) {
                try {
                	getNewMaxIdStatement.close();
                } catch (Exception e) {
                	e.printStackTrace();
                }
            }
        	if (connection != null) {
                try {
                	connection.close();
                } catch (Exception e) {
                	e.printStackTrace();
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.bc.xquarebc.listeners.NewDataStrategy#getNewData()
     */
    public Document getNewData() throws Exception {
        // formatting result message body :
        String formattedXquery = replaceDelimiterTokens(newDataXquery);
        logger.log(Level.FINE, "DelimiterNewDataStrategy formatted Xquery : " + formattedXquery);
        Document bodyDocument = processor.doQuery(formattedXquery, xqBridge);
        return bodyDocument;
    }

    protected String replaceDelimiterTokens(String sqlOrXquery) {
        sqlOrXquery = sqlOrXquery.replaceAll(XQUARE_PREVIOUS_DATA_DELIMITER_TOKEN, oldDataDelimiter);
        sqlOrXquery = sqlOrXquery.replaceAll(XQUARE_CURRENT_DATA_DELIMITER_TOKEN, newDataDelimiter);
        return sqlOrXquery;
    }
}
