/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.consume;

import java.io.File;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Properties;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;

import org.ow2.petals.bc.xquare.XQuareBC;
import org.ow2.petals.bc.xquare.XQuareBCException;
import org.ow2.petals.bc.xquare.XQuareConstants;
import org.ow2.petals.bc.xquare.util.XQuareUtils;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractExternalListener;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.component.framework.util.WSDLHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.Document;
import org.xquark.bridge.XQBridge;
import org.xquark.xml.xdbc.XMLConnection;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLResultSet;
import org.xquark.xml.xdbc.XMLStatement;

import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_DELIMITER_STRATEGY;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_POLLING_PERIOD;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_DATALISTENER_POLLING_PERIOD_DEFAULT;
import static org.ow2.petals.bc.xquare.XQuareConstants.XQUARE_VERIFY_TARGETED_SERVICE;

/**
 * <p>
 * This listener "listens" to new data in any XQuareBC service unit's target
 * database and builds upon the XQuery capabilities of the companion
 * XQuareBCJBIProcessor in order to get them, format them as XML and send it as
 * a message to the configured service. The "listening" part is done by
 * regularly looking for "new data" as the results of a configured SQL query.
 * </p>
 * 
 * <p>
 * WARNING : polling is not a best practice to integrate databases. Databases
 * are usually application-driven, and those applications (or their users)
 * should drive the "new data" lookup process. This implementation is provided
 * as fitting the most generic needs.
 * </p>
 * 
 * <h4>How it works</h4>
 * 
 * <p>
 * This default implementation identifies "new data" by a given field that must
 * be incremental - that is, be different and always increased in newly inserted
 * records. The usual case for it is a record identifier or primary key, but it
 * can work with other kind of fields as well, like timestamps. Concretely, it
 * executes the configured "oldDataDelimiterSelect" to get the old "new data
 * delimiting value", then it executes the configured "newDataDelimiterSelect"
 * to get the new "new data delimiting value" as available in the actual data.
 * If they are different, the configured XQuery is then executed through the
 * XQuareBCJBIProcessor and its results formatted as an XML document and sent as
 * a JBI message to the configured JBI service. Finally the "old data delimiting
 * value" is updated to the new one.
 * </p>
 * 
 * <p>
 * Notes :
 * </p>
 * This requires a new table to be added to the target database, that will
 * contain a single record with a single field being the last seen "maximum data
 * delimiting value". This table may start empty, in which case the
 * "initDataDelimiter" will be used the first time. <br/> Before executing the
 * XQuery, in its code the ##OLD_DATA_DELIMITER## qnd ##NEW_DATA_DELIMITER##
 * expressions are replaced by the values found in the preceding executions of
 * "oldDataDelimiterSelect" and "newDataDelimiterSelect". <br/> Many such
 * listeners may be deployed in different service units to listen to different
 * kinds of "new data" in the same database. <br/> Since it uses the
 * XQuareBCJBIProcessor's "query" capabilities, look at its documentation for
 * more details on how it behaves.
 * 
 * <h4>Configuration</h4>
 * <p>
 * The datasource to the target database and the XQuareBCJBIProcessor's XQuery
 * capabilities are configured through the XQuareBCJBIProcessor (look at its
 * javadoc for details).<br/> The configuration specific to this listener is
 * configured through the jbi.xml file provided by a given service
 * unit, and uses the following additional properties:
 * </p>
 * 
 * <ul>
 * <li>operation : This service's operation. Ex. sayHello</li>
 * <li>xquare.datalistener.polling.period : The period of the polling.
 * Optional (default value : 1000 ms).</li>
 * <li>xquare.verifyTargetedService: whether to verify the existence of
 * the recipient service and operation (search the WSDL of the service).
 * Possible values are true and false. Default value is false</li>
 * <li>xquare.dataListener.delimiterStrategy: The class name of the object who
 * detects new data. Two strategies are released with XQuare BC:
 * org.ow2.petals.bc.xquare.listeners.FileDelimiterNewDataStrategy and
 * org.ow2.petals.bc.xquare.listeners.DatabaseDelimiterNewDataStrategy.<br>
 * A custom class may be used. The class must extend
 * org.ow2.petals.bc.xquare.listeners.DelimiterNewDataStrategy
 * and be present in the Petals classpath.</li>
 * <li>xquare.datalistener.currentDelimiter : An SQL query that
 * must return the single value of the current "new data" delimiter. Ex. select
 * max(bid_date) from bids</li>
 * <li>xquare.datalistener.initialDelimiter : The data delimiter
 * initial value if none in its own table and field. Ex. 1970-01-01</li>
 * <li>xquare.datalistener.xquery : An XQuery that must return and
 * format the current new data. Note that the ##PREVIOUS_DATA_DELIMITER## and
 * ##CURRENT_DATA_DELIMITER## expressions are replaced within the SQL code before
 * its execution. Ex.:<br>
 * for $i in collection("bids")/bids<br>
 * where $i/bid_date &gt; xs:date("##PREVIOUS_DATA_DELIMITER##") and $i/bid_date
 * &lt;= xs:date("##CURRENT_DATA_DELIMITER##")<br>
 * return &lt;bid&gt;{ $i/bid_date }&lt;/bid&gt;</li>
 * </ul>
 * 
 * <p>Specific configuration for DatabaseDelimiterNewDataStrategy:</p>
 * 
 * <ul>
 * <li>xquare.datalistener.previousDelimiter : An SQL query that
 * must return the single value of the saved, old data delimiter. Ex. select
 * max(bid_date) from bids_xqbc</li>
 * <li>xquare.datalistener.updateDelimiters : An SQL insert that
 * must update the value of the "new data" delimiter from "previous" to "current". Note
 * that the ##CURRENT_DATA_DELIMITER## expression is replaced within the SQL code
 * before its execution. Ex.:<br>
 * delete from bids_xqbc;<br>
 * insert into bids_xqbc (bid_date) values ("##CURRENT_DATA_DELIMITER##")</li>
 * </ul>
 * 
 * <p>Specific configuration for FileDelimiterNewDataStrategy:</p>
 * 
 * <ul>
 * <li>xquare.datalistener.delimiterFolder (optional): The folder used to store the
 * previousDelimiter. The default value is {petals-folder}/xquare-bc-polling-files.</li>
 * <li>xquare.datalistener.delimiterFilename (optional): The filename used to store
 * the previousDelimiter. The default value is ServiceName.properties.</li>
 * </ul>
 * 
 * @version $Rev: $ $Date: $
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 * @author Roland NAUDIN - EBM Websourcing
 * 
 */
public class XQuareBCExternalListener extends AbstractExternalListener {

    private static Transformer transformer;

    private DatabaseQueryPoller poller;

    private boolean isInOut;

    protected QName service;

    protected String operation;
    
    private DelimiterNewDataStrategy newDataStrategy;

	/**
     * Get the transformer instance
     * 
     * @return the transformer instance
     * @throws TransformerConfigurationException
     */
    public synchronized static Transformer getTransformer()
            throws TransformerConfigurationException {
        if (transformer == null) {
            transformer = TransformerFactory.newInstance().newTransformer();
            Properties props = new Properties();
            props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperties(props);
        }
        return transformer;
    }

    /**
     * Sends an inOnly message
     * 
     * @param bodySource
     * @param service
     * @param operation
     * @return
     * @throws MessagingException
     * @throws PEtALSCDKException 
     */
    public boolean handleInOnlyMessage(Document bodyDocument) throws MessagingException, PEtALSCDKException {
        Exchange msg = createConsumeExchange();
        msg.setInMessageContent(bodyDocument);

        return sendSync(msg);
    }

    /**
     * Sends an inOut message
     * 
     * @param bodySource
     * @param service
     * @param operation
     * @return
     * @throws TransformerException
     * @throws MessagingException
     * @throws PEtALSCDKException 
     */
    public boolean handleInOutMessage(Document bodyDocument, QName service, String operation)
            throws MessagingException, TransformerException, PEtALSCDKException {

        Exchange msg = createConsumeExchange();
        msg.setInMessageContent(bodyDocument);

        sendSync(msg);

        // Verify if request was success,if not, handle the error message
        if (msg.getStatus().equals(ExchangeStatus.ERROR)) {
            // fault
            // TODO handle response !! output it as a file ??
            Source faultSource = msg.getFault().getContent();
            StringWriter buffer = new StringWriter();
            Result sResult = new StreamResult(buffer);
            getTransformer().transform(faultSource, sResult);
            getLogger().severe(
                    "XQuareBC data emission to service " + service + " and operation " + operation
                            + " received error response :" + buffer.toString());
        } else {
            // success
            // if there wasn't errors , binding component handle out reponse
            Source outSource = msg.getOutMessageContentAsSource();
            StringWriter buffer = new StringWriter();
            Result sResult = new StreamResult(buffer);
            getTransformer().transform(outSource, sResult);
            getLogger().finest(
                    "XQuareBC data emission " + "to service " + service + " and operation "
                            + operation + " received response :" + buffer.toString());
        }
        // Send status DONE to end exchange
        msg.setStatus(ExchangeStatus.DONE);
        send(msg);
        return true;

    }

    /**
     * Tries to find and process new data according to the configuration of the
     * given service unit.
     * 
     * TODO tries to extract data and send it first three params are the result
     * of the "process" target validation operation
     * 
     * @param inOut
     * @param service
     * @param operation
     * @param serviceProps
     * @throws Exception
     */
    public void handleNewData() throws Exception {
        String ref = XQuareUtils.generateKey(provides);
    	if (newDataStrategy == null) {
	        // getting a new data detection and handling strategy
            ConfigurationExtensions extensions = new ConfigurationExtensions(consumes.getAny());
	        String strategy = extensions.get(XQUARE_DATALISTENER_DELIMITER_STRATEGY);
	        XQBridge xqBridge = ((XQuareBC) this.component).getConsumerXQBridges().get(ref);
	        newDataStrategy = (DelimiterNewDataStrategy) Class.forName(strategy).getConstructor(new Class[] {}).newInstance();
	        newDataStrategy.init(this, xqBridge, extensions, getLogger());
        }
    	
        // detecting new data if any
        if (!newDataStrategy.detectNewData()) {
            // no data or no new data (when looping without having new data in
            // between) : do nothing
            return;
        }
        
        // querying for new data and getting the formatted new data message
        Document bodyDocument = newDataStrategy.getNewData();
        getLogger().finest(
                "Send request to service " + service + " and operation " + operation
                        + ". Sent data :" + XMLUtil.createStringFromDOMDocument(bodyDocument));

        boolean written = false;
        try {
            // TODO use AbstractComponent.sendMessage()
            if (!isInOut) {
                written = handleInOnlyMessage(bodyDocument);
            } else {
                // NB. in fact inOut and out very similar...
                written = handleInOutMessage(bodyDocument, service, operation);
            }
        } catch (MessagingException e) {
            getLogger().severe(e.getMessage());
        }
        if (written) {
            // updating known, "old" data
        	newDataStrategy.updateDataKnown();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.listener.AbstractExternalListener#start()
     */
    @Override
    protected void start() throws PEtALSCDKException {
        String errorMessage = new String();

        service = consumes.getServiceName();
        operation = consumes.getOperation();
        
        String verifyTargetedServiceValue = this.getExtensions().get(XQUARE_VERIFY_TARGETED_SERVICE);
        boolean verifyTargetedService = verifyTargetedServiceValue != null && verifyTargetedServiceValue.equals("true");
        
        if (verifyTargetedService && !verifyTargetedService(errorMessage)) {
            getLogger().severe("Service not found : " + errorMessage);
        } else {
            // retrieve the polling period
            String s_polling = getExtensions().get(XQUARE_DATALISTENER_POLLING_PERIOD);
            long period = XQUARE_DATALISTENER_POLLING_PERIOD_DEFAULT;
            if (!StringHelper.isNullOrEmpty(s_polling)) {
                period = Long.parseLong(s_polling);
                getLogger().finest("set polling period to :" + period + "ms");
            } else {
                getLogger().finest("set polling period to default :" + period + "ms");
            }
            poller = new DatabaseQueryPoller(period, 0, this);
            poller.startProcessing();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.listener.AbstractExternalListener#stop()
     */
    @Override
    protected void stop() throws PEtALSCDKException {
        poller.stopProcessing();
    }

    /**
     * Verify the existence and the matching of the targeted service.
     * 
     * @param errorMessage
     *            the message to fill if the validation fails
     * @return {@code true} if the verification success
     */
    private boolean verifyTargetedService(String errorMessage) {
        boolean found = false;

        // looking for valid endpoint
        errorMessage = "Looking for endpoints for service " + service + " with operation "
                + operation + " : ";
        ServiceEndpoint[] serviceEndpoints = this.component.getContext().getEndpointsForService(
                service);
        ServiceEndpoint serviceEndpoint = null;
        if (serviceEndpoints.length > 0) {
            // using first endpoint found
            serviceEndpoint = serviceEndpoints[0];
            Document serviceDesc;
            try {
                serviceDesc = this.component.getContext().getEndpointDescriptor(serviceEndpoint);

                // Determines if the operation exists and if it is an inOnly or
                // inOut operation
                try {
                    found = WSDLHelper.hasOperationNamed(serviceDesc, operation, service);
                    if (!found) {
                        errorMessage += "its WSDL definition (for endpoint "
                                + serviceEndpoint.getEndpointName() + ") has no such operation";
                    }
                    isInOut = WSDLHelper.isInOutOperation(serviceDesc, operation, service);
                } catch (Exception e) {
                    found = false;
                    isInOut = false;
                    errorMessage += "Error while looking in the wsdl definition "
                            + "for the required operation and its MEP for the endpoint "
                            + serviceEndpoint.getEndpointName();
                    getLogger().severe(errorMessage + " : " + e.getMessage());
                }

            } catch (JBIException e) {
                errorMessage += "Error getting descriptor of endpoint "
                        + serviceEndpoint.getEndpointName();
                getLogger().severe(errorMessage + " : " + e.getMessage());
            }
        } else {
            errorMessage += "No endpoint found";
        }

        return found;
    }

    /**
     * Implementation of the "query" operation.
     * 
     * @param content
     *            must be an XQuery consistent with the configured database and
     *            mapping of the specified service unit. It should be
     *            XML-encoded and wrapped in a "query" top tag (even if the
     *            parser tries to be nice by formatting and including XML
     *            subnodes).
     * @param bridge
     *            the service unit's bridge whose XML connection will be used to
     *            do the query
     * @param extensions
     *            service unit extensions
     * @return a Source corresponding of the formatted results of the XQuery,
     *         wrapped within the "Results" top tag.
     * @throws XQuareBCException
     */
    protected Document doQuery(String queryString, XQBridge bridge) throws XQuareBCException {
        if (queryString == null) {
            throw new XQuareBCException("No query found in a 'query' operation !");
        }

        // Getting the XML connection
        XMLConnection xc = null;
        try {
            xc = bridge.getXMLConnection();
        } catch (XMLDBCException e) {
            throw new XQuareBCException("XQuare XML DB error when getting XML connection", e);
        }
        
        String baseURI = this.getExtensions().get(XQuareConstants.XQUARE_XML_STATEMENT_BASE_URI);
        XMLStatement xs;
        try {
            // Creating the corresponding XML Statement
            xs = xc.createStatement();
            
            // set baseURI (should be absolute)
            if (baseURI == null || baseURI.trim().length() == 0 || !(new URI(baseURI).isAbsolute())) {
//                String suRootPath = ((XQuareSUManager) this.getComponent().getServiceUnitManager())
//                        .getSUInstallRootForActivatedEp(exchange.getEndpoint());
//                baseURI = "file://" + suRootPath;
                baseURI = new File("").toURI().toString();
            }
            xs.setBaseURI(baseURI);

            // executing the query
            XMLResultSet xrs = xs.executeQuery(queryString);
            // TODO line number limit, pagination

           Document resXmlDocument = XQuareUtils.convertToDocument(xrs);

            return resXmlDocument;
        } catch (URISyntaxException e) {
            throw new XQuareBCException("Incorrect base URI : " + baseURI, e);
        } catch (XMLDBCException e) {
            throw new XQuareBCException("XQuare XML DB error when trying to query", e);
        } finally {
            try {
                // properly closing the XML connection and the resources it
                // relies on
                xc.close();
            } catch (XMLDBCException e1) {
                this.getLogger().log(Level.WARNING, "Error closing XML connection");
            }
        }
    }
}
