/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ------------------------------------------------------------------------- $Id
 * : $ -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.su;

import java.io.File;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.sql.DataSource;

import org.ow2.petals.bc.xquare.XQuareBC;
import org.ow2.petals.bc.xquare.XQuareConstants;
import org.ow2.petals.bc.xquare.util.XQuareUtils;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.su.ServiceUnitListener;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.util.ClassLoaderUtil;
import org.ow2.petals.component.framework.util.StringHelper;
import org.xquark.bridge.Mapping;
import org.xquark.bridge.XQBridge;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 * <p>
 * Service Unit handler for the XQuare component.
 * </p>
 * 
 * <p>
 * At startup, lookups for a single *.wsdl file and a single *.properties file.
 * The first provides the service name, and the second provides properties that
 * are given to the XQuare component and from there to the XQuareBCJBIProcessor
 * and XQuareBCListener to use (look at their documentation about which
 * properties are available).
 * </p>
 * 
 * @version $Rev: 8655 $ $Date: 2008-07-17 16:13:01 +0200 (jeu., 17 juil. 2008)
 *  $
 * @since Petals 1.0
 * @author alouis
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 * 
 */
public class XQuareSUListener implements ServiceUnitListener {

    private Logger logger;

    private XQuareBC xQuareComponent;

    public XQuareSUListener(Logger logger, XQuareBC xQuareComponent) {
        this.logger = logger;
        this.xQuareComponent = xQuareComponent;
    }

    public void onSUDeployed(String serviceUnitName, String suRootPath, Jbi descriptor)
            throws PEtALSCDKException {

        // create a classLoader with SU libraries
        logger.finest("setup a classloader from service unit " + serviceUnitName + " :"
                + suRootPath);
        ClassLoader classLoader = ClassLoaderUtil.createClassLoader(suRootPath, getClass()
                .getClassLoader());

        // set xquare bridge for provides
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            String ref = XQuareUtils.generateKey(provides);

            ConfigurationExtensions extensions = null;
            if (this.xQuareComponent.getComponentConfiguration().getPropertiesFile() != null
                    && !StringHelper.isNullOrEmpty(this.xQuareComponent.getComponentConfiguration()
                            .getPropertiesFile().getValue())) {
                extensions = new ConfigurationExtensions(provides.getAny(), this.xQuareComponent
                        .getComponentConfiguration().getPropertiesFile().getValue());
            } else {
                extensions = new ConfigurationExtensions(provides.getAny());
            }

            // set datasource
            logger.finest("setup dataSource for " + ref);
            DataSource ds = setupDataSource(ref, extensions, classLoader);

            // FIXME : when the mapping is loaded on the xbridge,
            // queries no more works, even if the mapping is closed
            // so we use 2 different bridge objects

            // set xbridge for mapping
            logger.finest("setup an xqbridge for mapping for " + ref);
            XQBridge bridgeMapping = new XQBridge(ds);

            // set mapper
            logger.finest("setup mapping for " + ref);
            Mapping mapping = setupMapping(ref, bridgeMapping, extensions, ds, suRootPath);

            // set xbridge for queries
            logger.finest("setup an xqbridge for queries for " + ref);
            XQBridge bridgeQuery = new XQBridge(ds);

            // reference bridges and mapping, to be used by the JBIListener
            xQuareComponent.getProviderXQBridgesForMapping().put(ref, bridgeMapping);
            xQuareComponent.getXQBridgeMappings().put(bridgeMapping, mapping);
            xQuareComponent.getProviderXQBridgesForQuery().put(ref, bridgeQuery);

        }

        // set xquare bridge for consumes
        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {
            String ref = XQuareUtils.generateKey(consumes);

            ConfigurationExtensions extensions = null;
            if (this.xQuareComponent.getComponentConfiguration().getPropertiesFile() != null
                    && !StringHelper.isNullOrEmpty(this.xQuareComponent.getComponentConfiguration()
                            .getPropertiesFile().getValue())) {
                extensions = new ConfigurationExtensions(consumes.getAny(), this.xQuareComponent
                        .getComponentConfiguration().getPropertiesFile().getValue());
            } else {
                extensions = new ConfigurationExtensions(consumes.getAny());
            }

            // set datasource
            logger.finest("setup dataSource for " + ref);
            DataSource ds = setupDataSource(ref, extensions, classLoader);

            // set xbridge
            logger.finest("setup xqbridge for notifications for " + ref);
            XQBridge bridge = new XQBridge(ds);

            xQuareComponent.getConsumerXQBridges().put(ref, bridge);

        }
    }

    /**
     * @param ref
     * @param extensions
     * @return the datasource
     * @throws PEtALSCDKException
     */
    private DataSource setupDataSource(String ref, ConfigurationExtensions extensions,
            ClassLoader classLoader) throws PEtALSCDKException {
        DataSource ds = null;
        try {
            ds = xQuareComponent.getDataSources().getDataSource(extensions, classLoader, true);
            if (ds == null) {
                throw new Exception("DataSource for " + ref + " is null");
            }
        } catch (Exception e) {
            String msg = "Can not initialize datasource for " + ref;
            logger.log(Level.SEVERE, msg, e);
            throw new PEtALSCDKException(msg + e.getMessage());
        }
        return ds;
    }

    /**
     * setup a xqbridge mapping for consume/provide deployment
     * 
     * @param ref
     * @param extensions
     * @throws PEtALSCDKException
     */
    private Mapping setupMapping(String ref, XQBridge bridge, ConfigurationExtensions extensions,
            DataSource dataSource, String suRootPath) throws PEtALSCDKException {
        Mapping mapping = null;
        String mappingMode = extensions.get(XQuareConstants.XQUARE_DATABASE_MAPPING);
        if (mappingMode == null || mappingMode == "") {
            mappingMode = "[nomap]";
        }
        File databaseMappingFile = new File(suRootPath, mappingMode);
        String databaseMappingFilepath = databaseMappingFile.toURI().toString();

        // Getting the Mapping
        try {
            logger.finest("load mapping for " + ref);
            mapping = bridge.getMapping(databaseMappingFilepath);
        } catch (XMLDBCException e) {
            String msg = "XQuare : error when setting database mapping from file "
                    + databaseMappingFilepath;
            Throwable cause = e.getCause();
            if (cause == null) {
                cause = e.getException();
            }
            if (cause == null) {
                cause = e;
            }
            logger.log(Level.SEVERE, "Caused by : ", cause);
            throw new PEtALSCDKException(msg, cause);
        }
        return mapping;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.su.ServiceUnitListener#onSUStarted
     * (java.lang.String)
     */
    public void onSUStarted(String serviceUnitName) throws PEtALSCDKException {
    }

    /*
     * bridge.close(); (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.su.ServiceUnitListener#onSUStopped
     * (java.lang.String)
     */
    public void onSUStopped(String serviceUnitName) throws PEtALSCDKException {
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.su.ServiceUnitListener#onSUUndeployed
     * (java.lang.String, java.lang.String,
     * org.ow2.petals.jbi.descriptor.JBIDescriptor)
     */
    public void onSUUndeployed(String serviceUnitName, String suRootPath, Jbi descriptor)
            throws PEtALSCDKException {
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {

            ConfigurationExtensions extensions = null;
            if (this.xQuareComponent.getComponentConfiguration().getPropertiesFile() != null
                    && !StringHelper.isNullOrEmpty(this.xQuareComponent.getComponentConfiguration()
                            .getPropertiesFile().getValue())) {
                extensions = new ConfigurationExtensions(provides.getAny(), this.xQuareComponent
                        .getComponentConfiguration().getPropertiesFile().getValue());
            } else {
                extensions = new ConfigurationExtensions(provides.getAny());
            }

            String ref = XQuareUtils.generateKey(provides);
            logger.finest("remove bridges and mapping for " + ref);
            XQBridge bridge = xQuareComponent.getProviderXQBridgesForMapping().remove(ref);
            xQuareComponent.getXQBridgeMappings().remove(bridge);
            xQuareComponent.getProviderXQBridgesForQuery().remove(ref);
            // do not close either XBridge or mapping, because the DB
            // connections would be closed by the dataSources
            xQuareComponent.getDataSources().removeDataSource(extensions);

        }
        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {

            ConfigurationExtensions extensions = null;
            if (this.xQuareComponent.getComponentConfiguration().getPropertiesFile() != null
                    && !StringHelper.isNullOrEmpty(this.xQuareComponent.getComponentConfiguration()
                            .getPropertiesFile().getValue())) {
                extensions = new ConfigurationExtensions(consumes.getAny(), this.xQuareComponent
                        .getComponentConfiguration().getPropertiesFile().getValue());
            } else {
                extensions = new ConfigurationExtensions(consumes.getAny());
            }

            String ref = XQuareUtils.generateKey(consumes);
            logger.finest("remove bridge for notification for " + ref);
            xQuareComponent.getConsumerXQBridges().remove(ref);
            // do not close either XBridge, because DB connections would be closed
            // by the dataSources
            xQuareComponent.getDataSources().removeDataSource(extensions);
        }
    }
}
