/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ------------------------------------------------------------------------- $Id
 * : $ -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.util;

import java.sql.SQLException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.ow2.petals.bc.xquare.XQuareConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.xquark.jdbc.datasource.JDBCDataSource;
import org.xquark.jdbc.datasource.JDBCDataSourceFactory;

/**
 * 
 * @author Adrien LOUIS - EBMWebSourcing
 * @version $Rev: $ $Date:
 */
public class DataSources {

    private Map<String, DataSourceHandler> dataSources;

    private Logger logger;

    private class DataSourceHandler {

        public DataSourceHandler(DataSource dataSource) {
            this.dataSource = dataSource;
            this.holdCounter = 1;
        }

        public DataSource dataSource;

        public int holdCounter;
    }

    public DataSources(Logger logger) {
        this.logger = logger;
        dataSources = Collections.synchronizedMap(new HashMap<String, DataSourceHandler>());
    }

    /**
     * Return a dataSource related to the information stored in the extensions:
     * <li>jndiDataSource</li> <li>url, user, password</li> <br>
     * datasources are kept in a map, so their are created the first time, then
     * the existing instance is returned. <br>
     * datasources are shared for multiple service: if a service define a
     * connection already created, it is reused <br>
     * Connection properties are those from apache dbcp.
     * 
     * @see http://commons.apache.org/dbcp/configuration.html
     * @param extensions
     * @param classLoader
     *      classLoader that may contains extra jdbc drivers
     * @return the dataSource
     * @throws NamingException
     *      can not access jndi context
     */
    public DataSource getDataSource(ConfigurationExtensions extensions, ClassLoader classLoader,
            boolean onDeployed) throws SQLException, PEtALSCDKException, NamingException {
        DataSource dataSource = null;

        String jndiDataSource = extensions.get(XQuareConstants.XQUARE_JNDI_DATASOURCE);
        if (jndiDataSource != null) {
            // FIXME : test this feature
            logger.finest("FIXME: test the jndiDataSource loading feature");
            logger.finest("looking for jndiDataSource " + jndiDataSource + " in JNDI context");
            InitialContext jndiContext = new InitialContext();
            Object found = jndiContext.lookup(jndiDataSource);
            if (!(found instanceof DataSource)) {
                throw new NamingException("The JNDI object " + jndiDataSource
                        + " is not a DataSource object");
            }
            dataSource = (DataSource) found;
        } else {
            // look for a user defined data source
            String dsId = XQuareUtils.generateKey(extensions);
            DataSourceHandler dataSourceHandler = dataSources.get(dsId);
            if (dataSourceHandler != null) {
                dataSource = dataSourceHandler.dataSource;
                if (onDeployed) {
                    dataSourceHandler.holdCounter++;
                }                
                logger.finest("dataSource found in cache for " + dsId);
            } else {
                // NB. DB drivers beyond xquark's default four (see XQuark
                // JDBCDataSourceFactory) DON'T WORK !!!
                // creating a JDBC DataSource with XQuare's helper
                JDBCDataSourceFactory jdbcDataSourceFactory;
                jdbcDataSourceFactory = JDBCDataSourceFactory.newInstance(classLoader);

                // common properties
                String url = extensions.get(XQuareConstants.XQUARE_DATABASE_URL);
                String driver = extensions.get(XQuareConstants.XQUARE_DATABASE_DRIVER);

                String user = extensions.get(XQuareConstants.XQUARE_DATABASE_USER);
                String password = extensions.get(XQuareConstants.XQUARE_DATABASE_PASSWORD);

                // pool properties
                String maxActive = extensions.get(XQuareConstants.XQUARE_DATABASE_MAXACTIVE);
                String maxIdle = extensions.get(XQuareConstants.XQUARE_DATABASE_MAXIDLE);
                String minIdle = extensions.get(XQuareConstants.XQUARE_DATABASE_MINIDLE);
                String maxWait = extensions.get(XQuareConstants.XQUARE_DATABASE_MAXWAIT);
                String timeBEviction = extensions
                        .get(XQuareConstants.XQUARE_DATABASE_TIMEBTWEVICTION);

                logger.finest("create a dataSource with parameters :");
                logger.finest("-------------------------------------");
                logger.finest(dsId);

                dataSource = jdbcDataSourceFactory.newDataSource(driver, url, user, password);

                JDBCDataSource jdbcDataSource = (JDBCDataSource) dataSource;

                if (maxActive != null) {
                    jdbcDataSource.setMaxActive(Integer.parseInt(maxActive));
                    logger.finest("maxActive " + maxActive);
                }
                if (maxIdle != null) {
                    jdbcDataSource.setMaxIdle(Integer.parseInt(maxIdle));
                    logger.finest("maxIdle " + maxIdle);
                }
                if (minIdle != null) {
                    jdbcDataSource.setMinIdle(Integer.parseInt(minIdle));
                    logger.finest("minIdle " + minIdle);
                }
                if (maxWait != null) {
                    jdbcDataSource.setMaxWait(Integer.parseInt(maxWait));
                    logger.finest("maxWait " + maxWait);
                }
                if (timeBEviction != null) {
                    jdbcDataSource
                            .setTimeBetweenEvictionRunsMillis(Integer.parseInt(timeBEviction));
                    logger.finest("timeBetweenEvictionRunsMillis " + timeBEviction);
                }

                dataSources.put(dsId, new DataSourceHandler(dataSource));
            }
        }
        return dataSource;
    }

    public void removeDataSource(ConfigurationExtensions extensions) throws PEtALSCDKException {
        if (extensions.get(XQuareConstants.XQUARE_JNDI_DATASOURCE) == null) {
            String dsId = XQuareUtils.generateKey(extensions);
            DataSourceHandler dataSourceHandler = this.dataSources.get(dsId);
            dataSourceHandler.holdCounter--;
            if (dataSourceHandler.holdCounter == 0) {
                this.dataSources.remove(dsId);
                try {
                    ((JDBCDataSource) dataSourceHandler.dataSource).close();
                } catch (SQLException e) {
                    throw new PEtALSCDKException(e.getMessage());
                }
            }
        }
    }
}
