/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ------------------------------------------------------------------------- $Id
 * : $ -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.xquare.util;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.ow2.petals.bc.xquare.XQuareBCException;
import org.ow2.petals.bc.xquare.XQuareConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDocument;
import org.xquark.xml.xdbc.XMLResultSet;

/**
 * Utility class used by XQuare operations.
 * 
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 * @author Roland NAUDIN - EBM Websourcing
 * @author alouis - EBM Websourcing
 */
public class XQuareUtils {

    private final static ThreadLocal<DocumentBuilder> documentBuilder = new ThreadLocal<DocumentBuilder>() {

        @Override
        protected DocumentBuilder initialValue() {
            try {
                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                return factory.newDocumentBuilder();
            } catch (ParserConfigurationException e) {
                throw new RuntimeException("Failed to create Document builder", e);
            }
        }
    };

    /**
     * Get the content of the root element and control its local name.
     * 
     * @param document
     * @throws XQuareBCException
     *      The root element is not correct
     */
    public static final String getRootElementContent(Document document, String rootElementName)
            throws XQuareBCException {
        Element rootElement = document.getDocumentElement();
        if (!rootElementName.equals(rootElement.getLocalName())) {
            throw new XQuareBCException("XQuery statement(s) must have a root element named '"
                    + rootElementName + "'");
        }
        return rootElement.getTextContent();
    }

    /**
     * Converts the given XQuery results to XML and wrap them in a top "Results"
     * tag. NB. It generates the XML with the NameSpace defined in the
     * SQLXML_NAMESPACE_URI constant.
     * 
     * @param xrs
     *      results from an XQuery (or a stored one)
     * @return the corresponding, formatted XML Source
     * @throws XQuareBCException
     *      if any error
     */
    public static Document convertToDocument(XMLResultSet xrs) throws XQuareBCException {
        Document resultDocument = null;

        XMLDocument resXmlDoc;
        try {
            resXmlDoc = xrs.getFragmentsAsDocument(XQuareConstants.SQLXML_NAMESPACE_URI,
                    XQuareConstants.SQLXML_RESULTS_TAG, "xquare");

            resultDocument = resXmlDoc.getAsDocument();

        } catch (XMLDBCException e) {
            throw new XQuareBCException("Error while transforming result set into XML document", e);
        }

        return resultDocument;
    }

    /**
     * 
     * @param rs
     * @return
     * @throws ParserConfigurationException
     * @throws SQLException
     */
    public static final Document convertToRawDocument(ResultSet rs)
            throws ParserConfigurationException, SQLException {

        Document doc = documentBuilder.get().newDocument();

        Element results = doc.createElementNS(XQuareConstants.SQLXML_NAMESPACE_URI,
                XQuareConstants.SQLXML_RESULTS_TAG);
        doc.appendChild(results);

        ResultSetMetaData rsmd = rs.getMetaData();
        int colCount = rsmd.getColumnCount();

        while (rs.next()) {
            Element row = doc.createElement("Row");
            results.appendChild(row);

            for (int i = 1; i <= colCount; i++) {
                String columnName = rsmd.getColumnName(i);
                Object value = rs.getObject(i);

                Element node = doc.createElement(columnName);
                if (value != null) {
                    node.appendChild(doc.createTextNode(value.toString()));
                }
                row.appendChild(node);
            }
        }
        return doc;
    }

    /**
     * Return a string that is used for bridge and mapping hashMap storage.
     * Based on itf, srv, and ep
     * 
     * @param provides
     * @return
     */
    public static String generateKey(Provides provides) {
        return provides.getInterfaceName().toString() + "-" + provides.getServiceName() + "-"
                + provides.getEndpointName();
    }

    /**
     * Return a string that is used for bridge and mapping hashMap storage.
     * Based on itf, srv, and ep
     * 
     * @param consumes
     * @return
     */
    public static String generateKey(Consumes consumes) {
        return consumes.getInterfaceName().toString() + "-" + consumes.getServiceName() + "-"
                + consumes.getEndpointName();
    }

    /**
     * Return a string that is used for bridge and mapping hashMap storage.
     * Based on url,user,pwd and driver values
     * 
     * @param extensions
     * @return
     */
    public static final String generateKey(ConfigurationExtensions extensions) {
        String url = extensions.get(XQuareConstants.XQUARE_DATABASE_URL);
        String user = extensions.get(XQuareConstants.XQUARE_DATABASE_USER);
        String password = extensions.get(XQuareConstants.XQUARE_DATABASE_PASSWORD);
        String driver = extensions.get(XQuareConstants.XQUARE_DATABASE_DRIVER);
        return "driver:" + driver + " url:" + url + " usr:" + user + " pwd:" + password;
    }

}
